"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fuse = exports.FuseEffectSymbol = void 0;
const retil_support_1 = require("retil-support");
const observe_1 = require("./observe");
const source_1 = require("./source");
exports.FuseEffectSymbol = Symbol();
const throwArg = (error) => {
    throw error;
};
function fuse(fusor, onTeardown = retil_support_1.noop) {
    let onNext = null;
    let onError = throwArg;
    let onSeal = null;
    let onClear;
    let isFusing = false;
    let isInEffect = false;
    let isInvalidated = false;
    const used = new Map();
    const usedCores = new Map();
    const effectQueue = [];
    const effect = (callback) => {
        effectQueue.push(callback);
        return exports.FuseEffectSymbol;
    };
    const use = (source, ...defaultValues) => {
        const [core, select] = source;
        if (!used.has(core)) {
            used.set(core, []);
        }
        const usedState = { select, defaultValues };
        used.get(core).push(usedState);
        const usedCore = addCore(core);
        const result = usedCore.hasVersion
            ? select(usedCore.version)
            : defaultValues.length === 0
                ? select(core[0]())
                : defaultValues[0];
        usedState.result = result;
        return result;
    };
    const addCore = (core) => {
        let usedCore = usedCores.get(core);
        if (usedCore) {
            return usedCore;
        }
        else {
            const [getVersion, subscribe] = core;
            const change = () => {
                const usedStates = used.get(core);
                if (usedStates) {
                    const doesCoreHaveVersion = (usedCore.hasVersion = source_1.hasSnapshot([
                        core,
                    ]));
                    const coreVersion = (usedCore.version = usedCore.hasVersion
                        ? getVersion()
                        : undefined);
                    for (const usedState of usedStates) {
                        const hasDefaultValue = usedState.defaultValues.length;
                        const doesSourceHaveResult = doesCoreHaveVersion || hasDefaultValue;
                        const didSourceHaveResult = 'result' in usedState;
                        if (doesSourceHaveResult !== didSourceHaveResult ||
                            (didSourceHaveResult &&
                                usedState.result !==
                                    (doesCoreHaveVersion
                                        ? usedState.select(coreVersion)
                                        : usedState.defaultValues[0]))) {
                            isInvalidated = true;
                            if (!isInEffect && !isFusing) {
                                runFusor();
                            }
                            return;
                        }
                    }
                }
            };
            const seal = () => {
                usedCore.sealed = true;
                attemptSeal();
            };
            const initialHasVersion = source_1.hasSnapshot([core]);
            const usedCore = {
                hasVersion: initialHasVersion,
                version: initialHasVersion ? getVersion() : undefined,
            };
            usedCore.unsubscribe = subscribe(change, seal);
            usedCores.set(core, usedCore);
            return usedCore;
        }
    };
    const cleanUpUnusedCores = () => {
        for (const [core, { unsubscribe }] of Array.from(usedCores.entries())) {
            if (!used.has(core)) {
                usedCores.delete(core);
                unsubscribe();
            }
        }
        attemptSeal();
    };
    const attemptSeal = () => {
        if (!isFusing &&
            !isInvalidated &&
            !isInEffect &&
            !effectQueue.length &&
            onSeal &&
            (!used.size ||
                !Array.from(usedCores.values()).some((core) => !core.sealed))) {
            onSeal();
        }
    };
    const runEffects = () => {
        let result;
        isInEffect = true;
        const finishEffect = () => {
            isInEffect = false;
            if (isInvalidated) {
                runFusor();
            }
            else {
                cleanUpUnusedCores();
            }
            return result;
        };
        // Use an act to clear the result if the effects don't finish immediately,
        // and to keep the subscription open.
        act(() => {
            while (effectQueue.length) {
                const effect = effectQueue.shift();
                const effectResult = effect();
                result =
                    retil_support_1.isPromiseLike(result) || retil_support_1.isPromiseLike(effectResult)
                        ? Promise.resolve(result).then(() => effectResult)
                        : effectResult;
            }
            return retil_support_1.isPromiseLike(result)
                ? result.then(finishEffect, onError)
                : finishEffect();
        });
    };
    const runFusor = () => {
        if (!onNext || !onClear) {
            throw new Error('Retil error');
        }
        try {
            used.clear();
            isInvalidated = false;
            isFusing = true;
            const snapshot = fusor(use, effect);
            isFusing = false;
            if (snapshot === exports.FuseEffectSymbol) {
                runEffects();
            }
            else if (isInvalidated && effectQueue.length === 0) {
                runFusor();
            }
            else {
                onNext(snapshot);
                if (effectQueue.length) {
                    runEffects();
                }
                else {
                    cleanUpUnusedCores();
                }
            }
        }
        catch (errorOrPromise) {
            isFusing = false;
            if (retil_support_1.isPromiseLike(errorOrPromise)) {
                onClear();
                isInvalidated = true;
                errorOrPromise.then(() => {
                    // It's possible a source update has caused a run to complete in the
                    // intervening time.
                    if (onNext && isInvalidated && !isInEffect) {
                        runFusor();
                    }
                }, onError);
            }
            else {
                onError(errorOrPromise);
            }
        }
    };
    const source = observe_1.observe((next, error, seal, clear) => {
        onNext = next;
        onError = error;
        onSeal = seal;
        onClear = clear;
        isInvalidated = true;
        runFusor();
        return () => {
            onNext = null;
            onError = throwArg;
            onSeal = null;
            onClear = null;
            for (const { unsubscribe } of Array.from(usedCores.values())) {
                unsubscribe();
            }
            usedCores.clear();
            used.clear();
            onTeardown();
        };
    });
    const act = source[2];
    return source;
}
exports.fuse = fuse;
