"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.observe = exports.TEARDOWN_DELAY = void 0;
const retil_support_1 = require("retil-support");
const source_1 = require("./source");
exports.TEARDOWN_DELAY = 10;
function observe(observable) {
    const asyncActs = new Set();
    const callbacks = [];
    let actDeferred = null;
    let actDepth = 0;
    let error = null;
    let nextSnapshot = null;
    let sealed = false;
    let snapshot = null;
    let subscription = null;
    const observableSubscribe = typeof observable === 'function' ? observable : observable.subscribe;
    const get = () => {
        subscribeIfRequired();
        if (error) {
            throw error.value;
        }
        if (!snapshot) {
            snapshot = { deferred: new retil_support_1.Deferred() };
        }
        // This must be called *after* the snapshot is created, as it checks if
        // the snapshot is a deferred.
        scheduleTeardownIfRequired();
        if (!snapshot.deferred) {
            return snapshot.value;
        }
        // TODO: throw an object with a `then` that calls `subscribe`
        throw snapshot.deferred.promise;
    };
    const subscribe = (change, seal) => {
        if (sealed) {
            if (seal) {
                seal();
            }
            return retil_support_1.noop;
        }
        if (error) {
            return retil_support_1.noop;
        }
        const pair = [change, seal];
        callbacks.push(pair);
        subscribeIfRequired();
        return () => {
            const index = callbacks.indexOf(pair);
            if (index !== -1) {
                callbacks.splice(index, 1);
                scheduleTeardownIfRequired();
            }
        };
    };
    const handleSnapshot = (value) => {
        const hasValue = (snapshot && !snapshot.deferred) || nextSnapshot;
        const latestValue = nextSnapshot ? nextSnapshot.value : snapshot === null || snapshot === void 0 ? void 0 : snapshot.value;
        if (sealed || !subscription || (hasValue && latestValue === value)) {
            return;
        }
        nextSnapshot = { value };
        if (!actDepth) {
            commit();
        }
    };
    const handleClear = () => {
        nextSnapshot = null;
        if (sealed || !subscription || !snapshot || snapshot.deferred) {
            return;
        }
        snapshot = null;
        notifySubscribers();
    };
    const handleSeal = () => {
        if (!subscription) {
            return;
        }
        if (nextSnapshot) {
            commit();
        }
        if (snapshot === null || (snapshot === null || snapshot === void 0 ? void 0 : snapshot.deferred)) {
            handleError(new Error('Attempted to seal an observe() with no value'));
        }
        sealed = true;
        // Tear down the subscription *without* removing the value
        const unsubscribe = subscription.unsubscribe;
        subscription = null;
        try {
            unsubscribe();
        }
        catch (_a) { }
        callbacks.slice().forEach(([, seal]) => {
            if (seal) {
                seal();
            }
        });
        callbacks.length = 0;
    };
    const handleError = (err) => {
        if (sealed || !subscription) {
            return;
        }
        error = { value: err };
        const deferred = snapshot === null || snapshot === void 0 ? void 0 : snapshot.deferred;
        notifySubscribers();
        teardownSubscription();
        callbacks.length = 0;
        if (actDeferred) {
            actDeferred.reject(err);
        }
        if (deferred) {
            deferred.reject(err);
        }
    };
    const notifySubscribers = () => {
        // Some observables will immediately synchronously call `next` during
        // subscribe to let us know the current value. In this case, we'll
        // skip notifying subscribers, as they can get the value if they need
        // it.
        if (subscription && !subscription.isSubscribing) {
            callbacks.slice().forEach(callChangeListener);
        }
    };
    const subscribeIfRequired = () => {
        if (sealed) {
            return;
        }
        if (subscription) {
            subscription.count++;
            if (subscription.teardownTimeout) {
                clearTimeout(subscription.teardownTimeout);
            }
        }
        else if (!error) {
            subscription = {
                count: 1,
                isSubscribing: true,
            };
            const unsubscribeFunctionOrObject = observableSubscribe(handleSnapshot, handleError, handleSeal, handleClear);
            const unsubscribe = typeof unsubscribeFunctionOrObject === 'function'
                ? unsubscribeFunctionOrObject
                : unsubscribeFunctionOrObject.unsubscribe;
            if (subscription) {
                subscription.isSubscribing = false;
                subscription.unsubscribe = unsubscribe;
            }
            else {
                unsubscribe();
            }
        }
    };
    const scheduleTeardownIfRequired = () => {
        // Don't teardown if we've thrown a promises that hasn't yet resolved.
        if (subscription && --subscription.count === 0 && !(snapshot === null || snapshot === void 0 ? void 0 : snapshot.deferred)) {
            scheduleTeardown(subscription);
        }
    };
    const scheduleTeardown = (subscription) => {
        // TODO: use requestIdleCallback instead if possible.
        subscription.teardownTimeout = setTimeout(teardownSubscription, exports.TEARDOWN_DELAY);
    };
    const teardownSubscription = () => {
        // Avoid teardown if we've since resubscribed
        if (subscription && subscription.count === 0) {
            const unsubscribe = subscription.unsubscribe;
            nextSnapshot = null;
            snapshot = null;
            subscription = null;
            try {
                unsubscribe();
            }
            catch (_a) { }
        }
    };
    const commit = () => {
        if (sealed || !subscription || !nextSnapshot)
            return;
        const actDeferredCopy = actDeferred;
        const snapshotDeferred = snapshot === null || snapshot === void 0 ? void 0 : snapshot.deferred;
        const value = nextSnapshot.value;
        snapshot = nextSnapshot;
        nextSnapshot = null;
        actDeferred = null;
        // Some observables will immediately synchronously call `next` during
        // subscribe to let us know the current value. In this case, we'll
        // skip notifying subscribers, as they can get the value if they need
        // it.
        notifySubscribers();
        if (snapshotDeferred) {
            snapshotDeferred.resolve(value);
        }
        if (actDeferredCopy) {
            actDeferredCopy.resolve();
        }
        // If we skipped a teardown due to an unresolved promise, we can now finish
        // it off.
        if (subscription && subscription.count === 0) {
            scheduleTeardown(subscription);
        }
    };
    // TODO: queue until the first subscriber is added - include listeners to the
    // promise returned by this function as listeners. All callbacks become async
    // if there's no listeners, and require any value to be cleared.
    const act = (callback) => {
        if (error) {
            throw error.value;
        }
        if (sealed) {
            return Promise.resolve(callback());
        }
        const isTopLevelAct = ++actDepth === 1;
        const batch = (actDeferred = actDeferred || new retil_support_1.Deferred());
        subscribeIfRequired();
        const result = callback();
        if (retil_support_1.isPromiseLike(result)) {
            const asyncAct = result.then(() => {
                --actDepth;
                scheduleTeardownIfRequired();
                asyncActs.delete(asyncAct);
                if (asyncActs.size === 0) {
                    commit();
                }
            }, handleError);
            asyncActs.add(asyncAct);
            // Temporarily clear the result while waiting for the async action.
            if (snapshot && !(snapshot === null || snapshot === void 0 ? void 0 : snapshot.deferred)) {
                // Save the current snapshot in case nothing happens
                nextSnapshot = nextSnapshot || snapshot;
                snapshot = null;
                notifySubscribers();
            }
        }
        else {
            --actDepth;
            scheduleTeardownIfRequired();
            if (isTopLevelAct && asyncActs.size === 0) {
                commit();
            }
        }
        return batch.promise.then(() => result);
    };
    return [[get, subscribe], source_1.identitySelector, act];
}
exports.observe = observe;
const callChangeListener = ([listener]) => {
    try {
        if (listener) {
            listener();
        }
    }
    catch (errorOrPromise) {
        // Given callbacks will call `getSnapshot()`, which often throws a promise,
        // let's ignore thrown promises so that the callback don't have to.
        if (!retil_support_1.isPromiseLike(errorOrPromise)) {
            throw errorOrPromise;
        }
    }
};
